import { isRpcCallMessage, RPC_RESPONSE } from './common'
import type { AsyncFunction, RpcCallMessage, RpcResponseMessage } from './common'

export type { AsyncFunction }

/**
 * If the special symbol `injectRpcInfo` is passed as the first argument to a proxy function, this
 * argument will be replaced on the executing side by an `RpcInfo` object.
 */
 export interface RpcInfo {
	tab: browser.tabs.Tab,
}

/**
 * Register one or more functions to enable remote scripts to call them.
 *
 * @param functions - A `{ functionName: function }` mapping. Each function will be remotely
 * callable using the given name.
 */
export class RpcServer<Fs extends Record<string, AsyncFunction> = Record<string, AsyncFunction>> {
	constructor(
		public readonly functions: Fs,
	) {
		browser.runtime.onMessage.addListener(this.incomingRPCListener.bind(this))
	}

	// TODO Avoid conflict if there are multiple listeners.
	private incomingRPCListener(
		message: any,
		sender: browser.runtime.MessageSender,
	): undefined | Promise<RpcResponseMessage> {
		if (!isRpcCallMessage(message)) return

		// TODO Support extension popups and other pages, not just background script & tabs.
		// Each page gets the message, so we may need to name each endpoint.
		// Then here we should return if the message was not for us.

		return this.executeRpc(message, sender)
	}

	private async executeRpc(message: RpcCallMessage, sender: browser.runtime.MessageSender): Promise<RpcResponseMessage> {
		const funcName = message.funcName
		const func = this.functions[funcName]
		if (func === undefined) {
			console.error(`Received RPC for unknown function: ${funcName}`)
			return {
				rpcError: `No such function registered for RPC: ${funcName}`,
				__WEBEXTENSION_RPC_MESSAGE__: RPC_RESPONSE,
			}
		}

		const args = message.args

		if (message.addRpcInfoAsArgument !== false) {
			const rpcInfo: RpcInfo = {
				tab: sender.tab,
			}
			args[message.addRpcInfoAsArgument] = rpcInfo
		}

		// Run the function, return the result.
		try {
			const returnValue = await func(...args)
			return {
				returnValue,
				__WEBEXTENSION_RPC_MESSAGE__: RPC_RESPONSE,
			}
		} catch (error) {
			return {
				errorMessage: error.message,
				__WEBEXTENSION_RPC_MESSAGE__: RPC_RESPONSE,
			}
		}
	}
}
